//////////////////////////////////////////////////////////////////////////////////
//																				//
//	@author: kopffarben															//
//	@help: will create a MC														//
//	@tags: PRE ALPHA															//
//	@credits: vux																//
//																				//
//////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////
//																				//
//	OUTPUTBUFFER																//
//																				//
//////////////////////////////////////////////////////////////////////////////////
RWTexture3D<float4> RWDistanceVolume :  BACKBUFFER;

//////////////////////////////////////////////////////////////////////////////////
//																				//
//	INPUTS																		//
//																				//
//////////////////////////////////////////////////////////////////////////////////
cbuffer cbPerDraw : register( b0 )
{
	
	int PaintBufferCount	= 1;
	float3 VolumeDimension	= 1;
};
ByteAddressBuffer RawPositionSizeBuffer;
ByteAddressBuffer RawColorBuffer;

//////////////////////////////////////////////////////////////////////////////////
//																				//
//	CONST																		//
//																				//
//////////////////////////////////////////////////////////////////////////////////
static const uint groupDimCount = 512; // 512


//////////////////////////////////////////////////////////////////////////////////
//	CS SHARED RESOURCES															//
//	max 32Kb => float max count = 1024											//
//	max 32Kb => bool  maxCount for ARRAY 65536									//
//////////////////////////////////////////////////////////////////////////////////

groupshared float4	GroupPositionSizeSumHolder[groupDimCount];
groupshared uint	GroupBytePos = 0;
groupshared float4	GroupField = 0;
groupshared bool	GroupRepeat = true;


//////////////////////////////////////////////////////////////////////////////////
//																				//
//	CS SHADER FUNCTIONS	METABALLs												//
//																				//
//////////////////////////////////////////////////////////////////////////////////
float4 Metaball(float3 Pos, float4 PositionSize)
{
    float4 o;
	float RadiusSq = PositionSize.w;
    float3 d = (Pos/float3(VolumeDimension)) - PositionSize.xyz;
    float DistSq = dot(d, d);
    float InvDistSq = 1 / DistSq;
	
	o.xyz = -2 * RadiusSq * InvDistSq * InvDistSq * d;
    //o.w = RadiusSq * InvDistSq;
	 
	o.w =  InvDistSq *  RadiusSq* max((RadiusSq*4-DistSq),0);
	/*if (o.w != 0)
	{
		o.xyz = -2 * RadiusSq * InvDistSq * InvDistSq * d;
    } else {
    	o.xyz = 1;
    }*/

    return o;
}


//////////////////////////////////////////////////////////////////////////////////
//																				//
//	PASSES :)																	//
//																				//
//////////////////////////////////////////////////////////////////////////////////


void ReadPass(uint tid : SV_GroupIndex, uint3 groupIdx: SV_GroupID)
{
	uint BytePos = (GroupBytePos+tid);
	GroupPositionSizeSumHolder[tid] = 0;
	
	if (BytePos < (uint)PaintBufferCount)
	{
		GroupPositionSizeSumHolder[tid] += Metaball((float3)groupIdx.xyz , asfloat(RawPositionSizeBuffer.Load4(BytePos*16)));
	} 
	

	GroupMemoryBarrierWithGroupSync();	
}

void GroupSum(uint tid : SV_GroupIndex, uint pCount)
{
	if (groupDimCount >= pCount*2) 
	{

		uint BytePos = (GroupBytePos+tid);
		
		if (tid < pCount  && BytePos < (uint)PaintBufferCount) {
			GroupPositionSizeSumHolder[tid] += GroupPositionSizeSumHolder[tid + pCount];
		}

		GroupMemoryBarrierWithGroupSync();
	}
}
void GroupSumLast32(uint tid : SV_GroupIndex)
{	
	uint BytePos = (GroupBytePos+tid);
		
	if (tid < 32 && BytePos < (uint)PaintBufferCount) {
		if (groupDimCount >= 64) GroupPositionSizeSumHolder[tid] += GroupPositionSizeSumHolder[tid + 32];
		if (groupDimCount >= 32) GroupPositionSizeSumHolder[tid] += GroupPositionSizeSumHolder[tid + 16];
	}

	if (tid < 8 && BytePos < (uint)PaintBufferCount) {
		if (groupDimCount >= 16) GroupPositionSizeSumHolder[tid] += GroupPositionSizeSumHolder[tid + 8];
		if (groupDimCount >=  8) GroupPositionSizeSumHolder[tid] += GroupPositionSizeSumHolder[tid + 4];
	}

	if (tid < 4 && BytePos < (uint)PaintBufferCount) {
		if (groupDimCount >=  4) GroupPositionSizeSumHolder[tid] += GroupPositionSizeSumHolder[tid + 2];
		
	}
	if (tid < 2 && BytePos < (uint)PaintBufferCount) {
		if (groupDimCount >=  2) GroupPositionSizeSumHolder[tid] += GroupPositionSizeSumHolder[tid + 1];
	}

	if (tid == 0)
	{
		if (BytePos < (uint)PaintBufferCount)
		{
			GroupField += GroupPositionSizeSumHolder[0];
			GroupBytePos += groupDimCount;
			GroupRepeat = true;
		} else {
			GroupField += GroupPositionSizeSumHolder[0];
			GroupRepeat = false;
		}
	}
	GroupMemoryBarrierWithGroupSync();
}
void SumPass(uint tid : SV_GroupIndex, uint3 groupIdx: SV_GroupID)
{
	GroupSum(tid,512);
	GroupSum(tid,256);
	GroupSum(tid,128);
	GroupSumLast32(tid);
}
void WriteOutPass(uint tid : SV_GroupIndex, uint3 groupIdx: SV_GroupID)
{
		
	if (tid == 1 ) 
	{
		RWDistanceVolume[groupIdx] = GroupField;	
	}

}
//////////////////////////////////////////////////////////////////////////////////
//																				//
//	SANDBOX TESTING SHADER :)													//
//																				//
//////////////////////////////////////////////////////////////////////////////////
[numthreads(groupDimCount,1,1)]
void parallelSumCS(uint tid : SV_GroupIndex, uint3 groupIdx : SV_GroupID , uint3 groupTid: SV_GroupThreadID)
{
	bool repeat = true;
	while (repeat) 
	{
		ReadPass(tid,groupIdx);
		SumPass(tid,groupIdx);
		repeat = GroupRepeat;
	}
	WriteOutPass(tid,groupIdx);	
}
//////////////////////////////////////////////////////////////////////////////////
//																				//
//	TECHNIQUEs																	//
//																				//
//////////////////////////////////////////////////////////////////////////////////


technique11 parallelSum
{
	pass P0
	{
		SetComputeShader( CompileShader( cs_5_0, parallelSumCS() ) );
	}
}