using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel.Composition;
using System.Runtime.InteropServices;

using SlimDX;
using SlimDX.Direct3D11;

using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V1;

using FeralTic.DX11;
using FeralTic.DX11.Resources;
using SlimDX.DXGI;
using VVVV.DX11.Internals.Helpers;
using VVVV.DX11.Internals;
using FeralTic.DX11.Utils;

namespace VVVV.DX11.Nodes
{
    [PluginInfo(Name="IndexedGeometryTetrahedronsCube",Category="DX11.Geometry",Version="Join",Author="vux,tonfilm")]
    public class IndexedGeometryTetrahedronsCube : IPluginEvaluate, IDX11ResourceProvider, IDisposable
    {
        IPluginHost FHost;

        [Input("widthCount", DefaultValue = 1.0, Order = 0)]
		public ISpread<int> widthCount;
		
		[Input("heightCount", DefaultValue = 1.0, Order = 1)]
		public ISpread<int> heightCount;
		
		[Input("depthCount", DefaultValue = 1.0, Order = 2)]
		public ISpread<int> depthCount;

		[Input("from", DefaultValue = 0.0, Order = 3)]
		public ISpread<Vector3> origen;
		
		[Input("to", DefaultValue = 1.0, Order = 4)]
		public ISpread<Vector3> size;

        [Input("Apply",IsBang=true,DefaultValue=1 ,Order = 5)]
        protected ISpread<bool> FInApply;

        [Output("Geometry Out",IsSingle=true)]
        protected Pin<DX11Resource<DX11IndexedGeometry>> FOutput;

        bool FInvalidate;
        DataStream FVertexStream;
        DataStream FIndexStream;
        private int vertexsize;
        private InputElement[] inputlayout;
        private bool FFirst = true;

        [ImportingConstructor()]
        public IndexedGeometryTetrahedronsCube(IPluginHost host)
        {
            this.FHost = host;     
        }

    	private uint MakeIndicies(int x, int y, int z)
		{
			return (uint)(((z)*(heightCount[0] + 1) + (y))*(widthCount[0] + 1) + (x) /*+ baseIndex[0]*/);
		}
		
		private int VertexCount()
		{
			return ((depthCount[0]+1)*(heightCount[0]+1)*(widthCount[0]+1)*3);
		}
		
		private int IndexCount()
		{
			return (int)((depthCount[0])*(heightCount[0])*(widthCount[0])*24);
		}
    	
        public void Evaluate(int SpreadMax)
        {
            this.FInvalidate = false;
            if (this.FOutput[0] == null)
            {
                this.FOutput[0] = new DX11Resource<DX11IndexedGeometry>();
            }

            if (this.FInApply[0] || FFirst)
            {
                this.FFirst = false;
                this.FInvalidate = true;
                DX11Resource<DX11IndexedGeometry> instance = this.FOutput[0];
                this.FOutput[0] = instance;

                //Get Input Layout
                this.inputlayout = new InputElement[1];
                this.inputlayout[0] = new InputElement("POSITION", 0, Format.R32G32B32_Float, 0);
            	
            	this.vertexsize = 0;
                this.vertexsize += FormatHelper.Instance.GetSize(this.inputlayout[0].Format);
            	
                InputLayoutFactory.AutoIndex(this.inputlayout);
             
                //Load Vertex Stream
                if (this.FVertexStream != null) { this.FVertexStream.Dispose(); }

                this.FVertexStream = new DataStream(VertexCount() * Marshal.SizeOf(typeof(float)), true, true);
				this.FVertexStream.Position = 0;
            	
				float widthStep = size[0].X/widthCount[0];
				float heightStep = size[0].Y/heightCount[0];
				float depthStep = size[0].Z/depthCount[0];
				
				for (int z = 0; z <= depthCount[0]; z++) {
					for (int y = 0; y <= heightCount[0]; y++) {
						for (int x = 0; x <= widthCount[0]; x++) {
							this.FVertexStream.Write((float)(x*widthStep+origen[0].X));
							this.FVertexStream.Write((float)(y*heightStep+origen[0].Y));
							this.FVertexStream.Write((float)(z*depthStep+origen[0].Z));
						}
					}
				}
                this.FVertexStream.Position = 0;

                //Load index stream
                if (this.FIndexStream != null) { this.FIndexStream.Dispose(); }

                this.FIndexStream = new DataStream(IndexCount() * Marshal.SizeOf(typeof(uint)), true, true);
				this.FIndexStream.Position = 0;
            	
                for (int z = 0; z<depthCount[0]; z++) {
			        for (int y = 0; y<heightCount[0]; y++) {
			            for (int x = 0; x<widthCount[0]; x++) {
			                
			                // T0
			                this.FIndexStream.Write(MakeIndicies(x + 1, y, z));
							this.FIndexStream.Write(MakeIndicies(x, y, z));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y + 1, z));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y + 1, z + 1));
			
			                // T1
			                this.FIndexStream.Write(MakeIndicies(x + 1, y + 1, z + 1));
			                this.FIndexStream.Write(MakeIndicies(x, y, z));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y + 1, z));
			                this.FIndexStream.Write(MakeIndicies(x, y + 1, z));
			
			                // T2
			                this.FIndexStream.Write(MakeIndicies(x, y + 1, z));
			                this.FIndexStream.Write(MakeIndicies(x, y, z));
			                this.FIndexStream.Write(MakeIndicies(x, y + 1, z + 1));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y + 1, z + 1));
			
			                // T3
			                this.FIndexStream.Write(MakeIndicies(x, y, z));
			                this.FIndexStream.Write(MakeIndicies(x, y, z + 1));
			                this.FIndexStream.Write(MakeIndicies(x, y + 1, z + 1));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y + 1, z + 1));
			
			                // T4
			                this.FIndexStream.Write(MakeIndicies(x, y, z + 1));
			                this.FIndexStream.Write(MakeIndicies(x, y, z));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y, z + 1));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y + 1, z + 1));
			
			                // T5
			                this.FIndexStream.Write(MakeIndicies(x, y, z));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y, z));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y, z + 1));
			                this.FIndexStream.Write(MakeIndicies(x + 1, y + 1, z + 1));
						}
					}
				}
            	this.FIndexStream.Position = 0;

            }
        }

        #region IDX11Resource Members
        public void Update(IPluginIO pin, DX11RenderContext context)
        {
            if (this.FInvalidate)
            {
                if (this.FOutput[0].Contains(context)) { this.FOutput[0].Dispose(context); }

                DX11IndexedGeometry geom = new DX11IndexedGeometry(context);
                geom.InputLayout = this.inputlayout;
                geom.VertexSize = this.vertexsize;
  
                var vertices = new SlimDX.Direct3D11.Buffer(context.Device, this.FVertexStream, new BufferDescription()
                {
                    BindFlags = BindFlags.VertexBuffer,
                    CpuAccessFlags = CpuAccessFlags.None,
                    OptionFlags = ResourceOptionFlags.None,
                    SizeInBytes = (int)this.FVertexStream.Length,
                    Usage = ResourceUsage.Default
                });

                geom.HasBoundingBox = false;

                geom.IndexBuffer = new DX11IndexBuffer(context, this.FIndexStream, false, false);

                geom.InputLayout = this.inputlayout;
                geom.Topology = PrimitiveTopology.LineListWithAdjacency;

                geom.VertexBuffer = vertices;
                geom.VertexSize = this.vertexsize;
                geom.VerticesCount = VertexCount();

                this.FOutput[0][context] = geom;

                this.FInvalidate = false;
            }
        }

        public void Destroy(IPluginIO pin, DX11RenderContext context, bool force)
        {
            this.FOutput[0].Dispose(context);
        }
        #endregion

        #region IDisposable Members
        public void Dispose()
        {
            if (this.FOutput[0] != null) { this.FOutput[0].Dispose(); }
        }
        #endregion
    }
}
